import qrookDB.DB as qrdb
from data import QRTable
import unittest

# todo add tests for insert, update, delete, exec, function exec with returning

DB = qrdb.DB('postgres', 'qrookdb_test', 'kurush', 'pondoxo', format_type='dict')
op = DB.operators
DB.create_logger(app_name='qrookdb_test')
DB.create_data(__name__, in_module=True)

class TestBasic(unittest.TestCase):
    def test_tables_exist(self):
        self.assertTrue(isinstance(DB.a, QRTable))
        self.assertTrue(isinstance(DB.b, QRTable))
        self.assertTrue(isinstance(DB.ab, QRTable))

class TestRequest(unittest.TestCase):
    def assert_no_empty_data(self, data):
        self.assertTrue(data is not None and len(data) > 0)

    def assert_data_structure(self, data, type=None, rows_cnt=None, fields_cnt=None):
        if type:
            self.assertTrue(isinstance(data, type))
        if rows_cnt:
            self.assertTrue(data is not None and len(data) == rows_cnt)
        if fields_cnt:
            for d in data:
                self.assertTrue(len(d) == fields_cnt)

class TestSelect(TestRequest):
    def test_all(self):
        data1 = DB.select(DB.a).all()
        data2 = DB.a.select().all()
        self.assert_no_empty_data(data1)
        self.assert_no_empty_data(data2)

    def test_one(self):
        data1 = DB.select(DB.a).one()
        data2 = DB.b.select().one()
        self.assert_data_structure(data1, type=dict, rows_cnt=2)
        self.assert_data_structure(data2, type=dict, rows_cnt=2)

    def test_limit(self):
        data = DB.select(DB.a).limit(1).all()
        self.assert_data_structure(data, type=list, rows_cnt=1, fields_cnt=2)

    def test_offset(self):
        data = DB.select(DB.a, DB.a.id).order_by(DB.a.id).offset(5).all()
        for d in data:
            self.assertGreater(d['id'], 5)

    def test_order_by(self):
        data1 = DB.select(DB.a).order_by(DB.a.id).all()
        data2 = DB.select(DB.a).order_by(DB.a.id, desc=True).all()
        for i in range(1, len(data1)):
            self.assertGreater(data1[i]['id'], data1[i-1]['id'])
            self.assertLess(data2[i]['id'], data2[i-1]['id'])

    def test_distinct(self):
        data = DB.select(DB.ab, DB.ab.b_id, distinct=True).all()
        self.assert_data_structure(data, type=list, rows_cnt=3, fields_cnt=1)

    def test_add_attribute(self):
        query = DB.select(DB.a, DB.a.id)
        query.add_attribute(DB.a.name)
        data = query.all()
        self.assert_data_structure(data, type=list, fields_cnt=2)

    def test_group_by(self):
        data = DB.select(DB.a, 'count(*)').join(DB.ab, op.Eq(DB.a.id, DB.ab.a_id))\
            .join(DB.b, op.Eq(DB.b.id, DB.ab.b_id)).group_by(DB.b.info).all()

        data2 = DB.b.select().all()
        self.assert_data_structure(data, type=list, rows_cnt=len(data2), fields_cnt=1)
        pass

    def test_join(self):
        data = DB.select(DB.a, DB.a.id, DB.a.name, DB.b.id, DB.b.info).join(DB.ab, op.Eq(DB.a.id, DB.ab.a_id))\
            .join(DB.b, op.Eq(DB.b.id, DB.ab.b_id)).all()
        data2 = DB.ab.select().all()
        self.assert_data_structure(data, type=list, rows_cnt=len(data2), fields_cnt=4)

    def test_where(self):
        # todo test other ops
        data = DB.a.select().where(id=1).all()
        self.assert_data_structure(data, type=list, rows_cnt=1)

        data = DB.a.select().where(id=op.In(1,3,5,8)).all()
        self.assert_data_structure(data, type=list, rows_cnt=4)

        data = DB.a.select().where(id=1).where(id=2).all()
        self.assert_data_structure(data, type=list, rows_cnt=0)

        data = DB.a.select().where(id=1).where(id=2, bool='or').all()
        self.assert_data_structure(data, type=list, rows_cnt=2)

        data = DB.a.select().where(id=1).where(id=op.LE(10), name='Door', bool='or').all()
        self.assert_data_structure(data, type=list, rows_cnt=11)

    def test_errors(self):
        query = DB.a.select().where(id=1).where(id='a')
        data = query.all()
        self.assertEqual(data, None)
        self.assertEqual(query.get_error(), None)

        query = DB.a.select(DB.a.id, DB.a.id)
        data = query.all()
        self.assertEqual(data, None)
        self.assertTrue(isinstance(query.get_error(), str))

        query = DB.a.select(DB.a.id).where(id=1).add_attribute(DB.a.name)
        data = query.all()
        self.assertEqual(data, None)
        self.assertTrue(isinstance(query.get_error(), str))


# def main():
#
#     ok = DB.update(events, auto_commit=True).set(time=t).where(id=6).exec()
#     print(ok)
#
#     #ok = DB.insert(events, events.time, auto_commit=True).values([t]).exec()
#     #ok = DB.insert(events, events.date, events.time, auto_commit=True).values([d, t]).exec()
#     query = events.insert(events.date, events.time, auto_commit=False).values([[d, t], [None, t]]).returning('*')
#     data = query.all()
#     DB.commit()
#     print(data)
#
#     query = events.insert(events.date, events.time, auto_commit=True).values([[d, t], [None, t]]).returning(events.date, events.time)
#     data = query.all()
#     print(data)
#
#     query = events.insert(events.date, events.time, auto_commit=True).values([[d, t], [None, t]]).returning(['date', 'time'])
#     data = query.all()
#     print(data)
#
#     query = events.insert(events.date, events.time, auto_commit=True).values([[d, t], [None, t]]).returning('date', 'time')
#     data = query.all()
#     print(data)
#
#     data = DB.exec('select * from get_book_authors(1) as f(id int, name varchar)').config_fields('id', 'name').all()
#     print(data)
#
#     data = DB.funcall('get_book_authors', 1).config_fields('id', 'name').all()
#     print('funcall:', data)
#
#     q = DB.select(events, events.id, events.id).where(id=1)
#     data = q.all()
#     print('data:', data, ';\terror:', q.get_error())


if __name__ == '__main__':
    unittest.main()

sql = '''
create table A(id int primary key, name varchar);
create table B(id int primary key, info varchar);
create table AB(a_id int constraint a_fkey references A not null,
                b_id int constraint b_fkey references B not null);

insert into A values
(1, 'Kermit'), (2, 'Grover'),
(3, 'Сookie Monster'), (4, 'Beast'),
(5, 'Bert'), (6, 'Bertina'),
(7, 'Count von Count'), (8, 'Cranky'),
(9, 'Captain Cabbage'), (10, 'The Chief'),
(11, 'Doctor Two'), (12, 'Door'), (13, 'Elmo');

insert into B values
(1, 'fav character'), (2, 'odd character'), (3, 'aux character');

insert into AB values
(1, 2), (2, 2), (3, 1), (3, 2), (4, 1), (5, 3), (6, 3),
(7, 1), (8, 2), (9, 2), (10, 3), (11, 3),
(12, 3), (13, 2);
'''



import DI
import sys
from query import *
import operators
import data

# todo-list throughout the orm-project
'''
===important===
todo tests

===middle===
todo first connection - no logs, for logger not configured
todo insert values -> add array of dicts support
todo funcall - not working with psycorg2
todo connection via URL-string
todo select(a).join(...)... -> means 'select *', but only a-table attributes are returned

===future===
todo add support for nested queries
todo add select(books.id) parsing -> extract table from ONLY datafield
todo add returning part for delete query (and update?)
todo add StructDataFormatter
todo add 'as' syntax: 'select count(*) as cnt
'''

import qrlogging

class DBQueryAggregator:
    def __init__(self, conn: IConnector):
        self.connector = conn

    def exec(self, raw_query):
        return QRExec(self.connector, raw_query)

    def select(self, table: QRTable, *args, **kwargs):
        return QRSelect(self.connector, table, *args, **kwargs)

    def delete(self, table: QRTable, auto_commit=False, *args, **kwargs):
        return QRDelete(self.connector, table, auto_commit)

    def update(self, table: QRTable, auto_commit=False, *args, **kwargs):
        return QRUpdate(self.connector, table, auto_commit)

    def insert(self, table: QRTable, *args, auto_commit=False, **kwargs):
        return QRInsert(self.connector, table, *args, auto_commit=auto_commit)

    def funcall(self, func: str, *args, auto_commit=False, **kwargs):
        return QRFuncCall(self.connector, func, *args, auto_commit=auto_commit)

class DBCreator:
    """
    Class responsible for objects creation
    """

    def __init__(self, conn: IConnector, db: DBQueryAggregator):
        """
        :param conn: active connection for tables to use
        :param db: query aggregator
        """
        self.conn = conn
        self.db = db

    def create_data(self, source=None, in_module=False):
        """creates table data (described in DB.create_data) and returns it"""
        tables = self.conn.table_info()
        t = dict()

        for name, field in tables.items():
            t[name] = QRTable(name, field, self.db)

        self.__dict__.update(t)
        if source:
            if in_module:
                source = sys.modules[source]
            source.__dict__.update(t)
        return t


class DB:
    """facade class, recommended to be the only entry point for the whole orm
    (though it provides access to data such as operations and tables to be used separately)
    """
    operators = operators
    data = data

    def __init__(self, connector_type, *conn_args, format_type=None, **conn_kwargs):
        """
        :param connector_type: now only 'postgres' is supported
        :param conn_args: arg-params like host and port needed to connect to database
        :param format_type: the format data will be returned in; now only 'list' and 'dict' are supported
        :param conn_kwargs: kwarg-params like host and port needed to connect to database
        """
        DI.register(format_type, connector_type, *conn_args, **conn_kwargs)
        self.connector_type = connector_type
        self.conn_args = conn_args
        self.conn_kwargs = conn_kwargs
        self.format_type = format_type

        self.meta = dict()
        self.meta['connector'] = inject.instance(IConnector)
        self.meta['aggregator'] = DBQueryAggregator(self.meta['connector'])

    def __del__(self):
        self.meta['connector'].pool.closeall()

    def enable_database_drop(self) -> bool:
        return self.meta['connector'].enable_database_drop()

    def create_logger(self, logger_name='default', app_name='app',
                  level="INFO", file: str = None, file_level="INFO"):
        qrlogging.logger = qrlogging.create_logger(logger_name, app_name, level, file, file_level)

    def create_data(self, source=None, in_module=False):
        """ Create QRTable objects for all tables in given database (use system tables data)
            and stores these objects in this class.
            if 'source' present, attributes with names of found tables will be added to it
            'source' is object with __dict__ field or a module name (with 'in_module' flag up)
        """
        data = DBCreator(self.meta['connector'], self.meta['aggregator']).create_data(source, in_module)
        self.__dict__.update(data)

    def commit(self):
        self.meta['connector'].commit()

    def exec(self, raw_query):
        return self.meta['aggregator'].exec(raw_query)

    def select(self, table: QRTable, *args, **kwargs):
        return self.meta['aggregator'].select(table, *args, **kwargs)

    def delete(self, table: QRTable, auto_commit=False):
        return self.meta['aggregator'].delete(table, auto_commit)

    def update(self, table: QRTable, auto_commit=False):
        return self.meta['aggregator'].update(table, auto_commit)

    def insert(self, table: QRTable, *args, auto_commit=False):
        return self.meta['aggregator'].insert(table, *args, auto_commit=auto_commit)

    def funcall(self, func: str, *args, auto_commit=False, **kwargs):
        return self.meta['aggregator'].funcall(func, *args, auto_commit=auto_commit)



import psycopg2
import psycopg2.sql as sql
from psycopg2 import pool

from IConnector import IConnector, DBResult
from error_handlers import log_error, retry_log_error
from threading import Lock
import symbols
import qrlogging
import threading

# IConnector realization for Postgres database
class PostgresConnector(IConnector):
    def __init__(self, db, user, password, host='localhost', port=5432, schema='public', min_conn=1, max_conn=10):
        """
        configure connection
        :param db: db-name
        """
        self.db = db
        self.schema = schema
        self.user = user
        self.password = password
        self.host = host
        self.port = port
        self.connected = False
        self.min_conn = min_conn
        self.max_conn = max_conn

        self.enable_drop = False
        self.pool = None
        self.__connect()

    def __del__(self):
        if self.pool:
            self.pool.closeall()

    def enable_database_drop(self) -> bool:
        self.enable_drop = True
        return True

    @retry_log_error()
    def __connect(self):
        self.pool = psycopg2.pool.ThreadedConnectionPool(self.min_conn, self.max_conn, dbname=self.db, user=self.user,
                                     password=self.password, host=self.host, port=self.port)

    def exec(self, request: str, identifiers=None, literals=None, result='all'):
        conn = self.pool.getconn(key=threading.get_ident())
        cursor = conn.cursor()
        if self.enable_drop:
            conn.autocommit = True
            conn.set_isolation_level(0)
        #request = request.replace(symbols.QRDB_IDENTIFIER, '{}')
        request = request.replace(symbols.QRDB_LITERAL, '%s')
        if identifiers:
            #identifiers = [sql.Identifier(x) for x in identifiers]
            identifiers = ['"' + x + '"' for x in identifiers]
            i = 0
            while request.find(symbols.QRDB_IDENTIFIER) != -1:
                request = request.replace(symbols.QRDB_IDENTIFIER, identifiers[i], 1)
                i += 1
            if len(identifiers) != i:
                raise Exception('unexpected identifiers count')
            #request = request.format(*identifiers)

        request = sql.SQL(request)
        qrlogging.info('POSTGRES EXECUTE: %s with literals %s', request.as_string(cursor), literals)
        try:
            #cursor.execute('''SET search_path TO %s, public;''' % self.schema)
            cursor.execute(request, literals)
        except Exception as e:
            conn.rollback()
            raise e
        data = self.extract_result(cursor, result)
        cursor.close()
        #if request.startswith('select'):
        if conn.autocommit:
            self.pool.putconn(conn, key=threading.get_ident())
        return DBResult(data, result)

    def extract_result(self, cursor, result):
        if result == 'all':
            return cursor.fetchall()
        elif result == 'one':
            return cursor.fetchone()
        elif result is not None:
            qrlogging.warning("unexpected 'result' value: %s" % result)
        return None

    def table_info(self):
        request = '''select distinct table_name, column_name, data_type
                     from information_schema.columns
                     where table_schema = '%s';''' % self.schema

        data = self.exec(request, result='all')
        info = {}
        for d in data.get_data():
            if not info.get(d[0]):
                info[d[0]] = []
            info[d[0]].append((d[1], d[2]))
        return info

    def commit(self):
        conn = self.pool.getconn(key=threading.get_ident())
        conn.commit()
        self.pool.putconn(conn, key=threading.get_ident())



import setuptools

setuptools.setup(
    name="qrookDB",
    version="1.3.2.1",
    author="Kurush",
    author_email="ze17@ya.ru",
    description="tiny ORM for SQL-databases",
    long_description_content_type="text/markdown",
    url="https://github.com/Kurush7/qrook_db",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    long_description='''
    
# Initializing
This package represents a new ORM to work with SQL-syntax databases (for now, only PostgreSQL and SQLite are natively supported,
but you may inherit your own connectors from IConnector abstract class).

To start working, all you need to do is import a package's facade object and initialise it:
```python
import qrookDB.DB as DB
DB = db.DB('postgres', 'db_name', 'username', 'password', format_type='dict')
DB.create_logger(app_name='qrookdb_test', file='app_log.log')
DB.create_data(__name__, in_module=True)
```
Here we first created a database connection, providing connect parameters (format_type defines the form
in which results'll be returned - 'list' and 'dict' supported),
then initialized an internal logger to write both in console and file.
Note that instead of 'postgres' you could've sent the instance of the connector (including your own connectors).
The 'create_data' function reads database system tables to get info about all user-defined tables,
and creates QRTable objects based on this info. Now you can use table names (ones given to them in database)
to access to these objects as DB instance fields (also, configuration showed above adds these table names to
your current module, so you can use short names: 'books' instead of 'DB.books').   

# Querying
You can form execute queries using either DB instance or concrete tables (in this case, you don't need to
mention in which table to perform queries). To execute any query, use one of 'exec', 'one' and 'all' methods
(latter two define how many rows to return from query; 'exec' returns None by default). If forming of the query
fails, it won't be executed at all (and will return None if you try); you can use 'get_error' method of the query
to get the error description (it will be logged, though). Note: if error occured in the middle of query-building,
query will ignore the rest of building proccess.
## Select queries examples
```python
op = DB.operators

print(DB.books)
print(books, books.id)

# logical 'and' is used by default for multiple where conditions; 'op' module contains special operators
data = DB.select(books).where(original_publication_year=2000, language_code='eng').\
    where(id=op.In(470, 490, 485)).all()

# you can add raw-string query parts, but it'll be on your conscience in terms of security   
query = books.select('count(*)').group_by(books.original_publication_year)
data = query.all()

data = DB.select(books, books.id).where('id < 10').order_by(books.id, desc=True).\
    limit(3).offset(2).all()

# here fields have same name ('id'), but via different tables it'll be ok
# (for data in dict-format, table-names'll be added to keys) 
data = books.select(authors.id, books.id)\
    .join(books_authors, op.Eq(books_authors.book_id, books.id))\
    .join(authors, op.Eq(books_authors.author_id, authors.id)).all()
    # .join(books_authors, 'books_authors.book_id = books.id')\

data = books.select(books.id).where(id=1).where(bool='or', id=2).all()

# error - trying to select two equal fields;
q = DB.select(events, events.id, events.id).where(id=1)
data = q.all()
print('data is None here:', data, ';\terror:', q.get_error())
```


## Update, Insert, Delete queries examples
```python
# if auto_commit is not set, you'll have to commit manually 
ok = DB.delete(events, auto_commit=True).where(id=1).exec()

from datetime import datetime
t = datetime.now().time()
d = datetime.now().date()
ok = DB.update(events, auto_commit=False).set(time=t).where(id=6).exec()
DB.commit()

# other possible variants for values: values([t]), values([d, t])
# other possible variants for returning: returning(events.date, events.time), returning(['date', 'time']), returning('date', 'time')
query = events.insert(events.date, events.time, auto_commit=True).values([[d, t], [None, t]]).returning('*')
data = query.all()

# you can also execute fully-raw queries; if you need to return values, 
use 'config_fields' to define results' names (not necessary for 'list' data format) 
data = DB.exec('select * from get_book_authors(1) as f(id int, name varchar)').config_fields('id', 'name').all()
print(data)
```
'''
)
